@@ -38,19 +38,22 @@ from hypothesis.core import find
 from hypothesis.errors import Flaky, NoSuchExample, InvalidDefinition, \
     HypothesisException
 from hypothesis.control import BuildContext
-from hypothesis._settings import settings as Settings
 from hypothesis._settings import Verbosity
+from hypothesis._settings import settings as Settings
 from hypothesis.reporting import report, verbose_report, current_verbosity
 from hypothesis.strategies import just, lists, builds, one_of, runner, \
     integers
 from hypothesis.vendor.pretty import CUnicodeIO, RepresentationPrinter
 from hypothesis.internal.reflection import proxies, nicerepr
 from hypothesis.internal.conjecture.data import StopTest
-from hypothesis.internal.conjecture.utils import integer_range
+from hypothesis.internal.conjecture.utils import integer_range, \
+    calc_label_from_name
 from hypothesis.searchstrategy.strategies import SearchStrategy
 from hypothesis.searchstrategy.collections import TupleStrategy, \
     FixedKeysDictStrategy
 
+STATE_MACHINE_RUN_LABEL = calc_label_from_name('another state machine step')
+
 
 class TestCaseProperty(object):  # pragma: no cover
 
@@ -116,7 +119,6 @@ def run_state_machine_as_test(state_machine_factory, settings=None):
 
 
 class GenericStateMachine(object):
-
     """A GenericStateMachine is the basic entry point into Hypothesis's
     approach to stateful testing.
 
@@ -211,7 +213,6 @@ GenericStateMachine.find_breaking_runner = classmethod(find_breaking_runner)
 
 
 class StateMachineRunner(object):
-
     """A StateMachineRunner is a description of how to run a state machine.
 
     It contains values that it will use to shape the examples.
@@ -236,7 +237,7 @@ class StateMachineRunner(object):
             while True:
                 if steps >= self.n_steps:
                     stopping_value = 0
-                self.data.start_example()
+                self.data.start_example(STATE_MACHINE_RUN_LABEL)
                 if not cu.biased_coin(self.data, stopping_value):
                     self.data.stop_example()
                     break
@@ -430,7 +431,6 @@ class ShuffleBundle(object):
 
 
 class RuleBasedStateMachine(GenericStateMachine):
-
     """A RuleBasedStateMachine gives you a more structured way to define state
     machines.
 
@@ -441,6 +441,7 @@ class RuleBasedStateMachine(GenericStateMachine):
     executed.
 
     """
+
     _rules_per_class = {}
     _invariants_per_class = {}
     _base_rules_per_class = {}
